import os
import subprocess
import tempfile

import pytest

def test_changes():
    """
    Confirm that the consumer end of the transmission stays in-sync with the producer end over a series of
    changes to the producer-end of the filesystem. We are going to do this by running the producer and the
    server as subprocesses and then interacting with the server through http requests.
    """
    server_process = subprocess.run(
        f"python {dot_slash('server.py')}",
        shell=True,
        capture_output=True,
    )



def dot_slash(relpath):
    """
    Creates an absolute path from the specified relative path with the packages root 
    as the base of the path.
    """
    here = os.path.dirname(os.path.realpath(__file__))
    return os.path.normpath(os.path.join(here, relpath))


import base64
from enum import Enum
import json

from pydantic import BaseModel

class EventAction(Enum):
    CREATE="CREATE"
    DELETE="DELETE"
    UPDATE="UPDATE"

class ChannelEvent(BaseModel):
    filepath: str
    action: EventAction
    contents: bytes
    nonce: bytes

    def to_dict(self) -> dict:
        return {
            "filepath": self.filepath,
            "action": self.action.value,
            "contents": base64.b64encode(self.contents).decode(),
            "nonce": base64.b64encode(self.nonce).decode(),
        }

    @classmethod
    def from_dict(cls, d: dict):
        action: EventAction
        raw_action = d["action"]
        if raw_action == "CREATE":
            action = EventAction.CREATE
        elif raw_action == "DELETE":
            action = EventAction.DELETE
        elif raw_action == "UPDATE":
            action = EventAction.UPDATE
        else:
            raise ValueError(f"Bad value for action: {raw_action}")
        return cls(
            filepath=d["filepath"],
            action=action,
            contents=base64.b64decode(d["contents"].encode()),
            nonce=base64.b64decode(d["nonce"].encode()),
        )

class FileState(BaseModel):
    filepath: str
    contents: bytes
    nonce: bytes

    @classmethod
    def from_channel_event(cls, e: ChannelEvent) -> "FileState":
        return cls(
            filepath=e.filepath,
            contents=e.contents,
            nonce=e.nonce,
        )

    def to_dict(self) -> dict:
        return {
            "filepath": self.filepath,
            "contents": base64.b64encode(self.contents).decode(),
            "nonce": base64.b64encode(self.nonce).decode(),
        }

    @classmethod
    def from_dict(cls, d: dict):
        return cls(
            filepath=d["filepath"],
            contents=base64.b64decode(d["contents"].encode()),
            nonce=base64.b64decode(d["nonce"].encode()),
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="holographic",
    version="0.0.1",
    author="Max Taggart",
    author_email="max.taggart@healthcatalyst.com",
    description="This is ground control to Major Tom...",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "click~=8.1.3",
        "cryptography~=39.0.0",
        "Flask~=2.2.2",
        "gunicorn~=20.1.0",
        "pydantic~=1.10.2",
        "requests~=2.28.1",
        "rich~=13.0.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import Dict, Iterable, List
import uuid

from flask import Flask, request

from .models import (
    ChannelEvent,
    EventAction,
    FileState,
)


class ClientError(Exception):
    def __init__(self, message: str):
        self.message = message

class NotFoundError(Exception):
    def __init__(self, message: str):
        self.message = message

class Channel:
    def __init__(self, id: str):
        self.id = id
        self.all_files: Dict[str, FileState] = {}
        self.updates: Dict[str, ChannelEvent] = {}

    def __repr__(self) -> str:
        num_files = len(self.all_files)
        unretrieved_updates = len(self.updates)
        return f"<Channel ({self.id}) files: {num_files}  updates: {unretrieved_updates}>"

    def new_events(self, events: Iterable[ChannelEvent]) -> None:
        for e in events:
            if e.action in [EventAction.CREATE, EventAction.UPDATE]:
                self.all_files[e.filepath] = FileState.from_channel_event(e)
                if e.filepath in self.updates:
                    old_event = self.updates[e.filepath]
                    old_event.action = EventAction.UPDATE
                    old_event.contents = e.contents
                    old_event.nonce = e.nonce
                else:
                    self.updates[e.filepath] = e
            if e.action == EventAction.DELETE:
                if e.filepath in self.all_files:
                    del self.all_files[e.filepath]
                if e.filepath in self.updates:
                    old_event = self.updates[e.filepath]
                    old_event.action = EventAction.DELETE
                    old_event.contents = e.contents
                    old_event.nonce = e.nonce

    def get_updates(self) -> List[ChannelEvent]:
        updates = list(self.updates.values())
        self.updates = {}
        return updates

    def get_all_files(self) -> List[FileState]:
        return list(self.all_files.values())

class EventStore:
    def __init__(self):
        # Holds the events that haven't been retrieved yet
        self.channels: Dict[str, Channel] = {}

    def new_channel(self):
        channel_id = str(uuid.uuid4())
        self.channels[channel_id] = Channel(channel_id)
        return channel_id

    def update_channel(self, channel_id: str, events: Iterable[ChannelEvent]):
        if channel_id not in self.channels:
            raise NotFoundError(f"There is no channel with id '{channel_id}'.")
        self.channels[channel_id].new_events(events)
        
    def get_channel_updates(self, channel_id: str) -> List[ChannelEvent]:
        if channel_id not in self.channels:
            raise NotFoundError(f"There is no channel with id '{channel_id}'.")
        return self.channels[channel_id].get_updates()

    def get_all_files_from_channel(self, channel_id: str) -> List[FileState]:
        if channel_id not in self.channels:
            raise NotFoundError(f"There is no channel with id '{channel_id}'.")
        return self.channels[channel_id].get_all_files()

    def close_channel(self, channel_id: str):
        if channel_id not in self.channels:
            raise NotFoundError(f"There is no channel with id '{channel_id}'.")
        del self.channels[channel_id]


app = Flask(__name__)
event_store = EventStore()

@app.get("/channel/<channel_id>/updates")
def get_updates_route(channel_id):
   return get_channel_updates(channel_id)

@app.route("/channel/<channel_id>", methods=["GET", "POST", "DELETE"])
def channel_route(channel_id):
    try:
        if request.method == "GET":
            return get_channel_state(channel_id)
        elif request.method == "POST":
            return update_channel(channel_id)
        elif request.method == "DELETE":
            return close_channel(channel_id)
        else:
            return "Method not allowed", 405
    except ClientError as e:
        return e.message, 400
    except NotFoundError as e:
        return e.message, 404

@app.post("/channel")
def new_channel_route():
    return new_channel()

def new_channel():
    global event_store
    channel_id = event_store.new_channel()
    return channel_id

def close_channel(channel_id):
    global event_store
    event_store.close_channel(channel_id)
    return ""

def update_channel(channel_id):
    global event_store
    body_json = request.json
    if body_json is None:
        return "Body was not valid JSON.", 400
    raw_events: List[dict] = body_json
    events = [ChannelEvent.from_dict(e) for e in raw_events]
    event_store.update_channel(channel_id, events)
    return ""

def get_channel_state(channel_id) -> List[dict]:
    global event_store
    files = [f.to_dict() for f in event_store.get_all_files_from_channel(channel_id)]
    return files

def get_channel_updates(channel_id) -> List[dict]:
    global event_store
    updates = [u.to_dict() for u in event_store.get_channel_updates(channel_id)]
    return updates


if __name__ == "__main__":
    app.run(host="127.0.0.1", port=7788, debug=True)


import base64
import os
from pathlib import Path
import textwrap
import time
import secrets
from typing import (
    Tuple,
    List,
    Union,
    Dict,
)

import click
from cryptography.hazmat.primitives.ciphers.aead import AESOCB3
from pydantic import BaseModel
import requests
from rich.text import Text
from rich.console import Console
from rich.syntax import Syntax

from .models import (
    ChannelEvent,
    EventAction,
)


class FileAdd(BaseModel):
    path: str
    contents: str

    def to_channel_event(self, encryptor: AESOCB3) -> ChannelEvent:
        contents, nonce = encrypt_contents(encryptor, self.contents.encode(), self.path.encode())
        return ChannelEvent(
            filepath=self.path,
            action=EventAction.CREATE,
            contents=contents,
            nonce=nonce,
        )

class FileDelete(BaseModel):
    path: str

    def to_channel_event(self, encryptor: AESOCB3) -> ChannelEvent:
        # Generate some cryptographically random contents so that we
        # don't use an empty string every time for delete actions. I suspect
        # that using the same contents every time for delete actions
        # may make it easier to guess the key given that the cyphertext
        # will then only be derived from the key, the nonce, and the associated data,
        # where both the nonce and the associated data are public.
        random_contents = secrets.token_bytes(128)
        contents, nonce = encrypt_contents(encryptor, random_contents, self.path.encode())
        return ChannelEvent(
            filepath=self.path,
            action=EventAction.DELETE,
            contents=contents,
            nonce=nonce,
        )


class FileUpdate(BaseModel):
    path: str
    contents: str

    def to_channel_event(self, encryptor: AESOCB3) -> ChannelEvent:
        contents, nonce = encrypt_contents(encryptor, self.contents.encode(), self.path.encode())
        return ChannelEvent(
            filepath=self.path,
            action=EventAction.UPDATE,
            contents=contents,
            nonce=nonce,
        )


def encrypt_contents(encryptor: AESOCB3, contents: bytes, associated_data: bytes) -> Tuple[bytes, bytes]:
    """
    Encrypts the contents and returns a tuple of <encrypted_contents> and
    <nonce>.
    """
    nonce = os.urandom(12)
    cyphertext = encryptor.encrypt(nonce, contents, associated_data)
    return cyphertext, nonce


class File(BaseModel):
    path: str
    last_modified: float
    contents: str


# http://127.0.0.1:7788

@click.command()
@click.argument("url")
@click.argument("target_dir")
def main(url, target_dir):
    console = Console()
    channel_id = ""
    try:
        channel_id = open_channel(url)
        watch_and_emit(url, channel_id, target_dir)
    except KeyboardInterrupt:
        if channel_id != "":
            close_channel(url, channel_id)
        console.print("\nTransmission ended.", style="bold")
        

def watch_and_emit(url: str, channel_id: str, target_dir_arg: str):
    target_dir = Path(target_dir_arg)
    if not target_dir.exists():
        raise ValueError(f"Direcotry '{str(target_dir)}' does not exist.")
    target_dir = os.path.abspath(target_dir) + "/"
    # Instantiate the encryption machinery
    key = AESOCB3.generate_key(bit_length=256)
    encryptor = AESOCB3(key)
    print_channel_details(url, channel_id, key)
    poll_interval_ms = 750
    old_files_set = set()
    files_map: Dict[str, File] = {}
    while True:
        events: List[Union[FileAdd, FileUpdate, FileDelete]] = []
        new_files_set = set()
        for dirpath, _, filenames in os.walk(target_dir):
            for filename in filenames:
                abs_path = os.path.abspath(os.path.join(dirpath, filename))
                # In the destination filesystem we want to write the contents of `target_dir`, not the
                # directory its self or any of its ancestors. So to get the filepath as it should be
                # consumed by the remote system we need to remove the path to the target directory from
                # the paths of the files in the directory. The resulting path is `holograph_path`. e.g.
                # if there is a file in target_dir whose absolute path is 
                # '/user/grandparent/parent/target_dir/package1/module1.py', this file will be transmitted
                # to the server with the path 'package1/module1.py'.
                holograph_path = abs_path.replace(target_dir, "")
                new_files_set.add(holograph_path)
                # Check if this file existed last time we polled
                if holograph_path in files_map:
                    old_file_info = files_map[holograph_path]
                    # it did
                    # check if it's been updated
                    last_updated = os.path.getmtime(abs_path)
                    if last_updated == old_file_info.last_modified:
                        # nope
                        continue
                    # yep, update the info in the files_map
                    with open(abs_path) as h:
                        contents = h.read()
                    files_map[holograph_path] = File(
                        path=holograph_path,
                        last_modified=last_updated,
                        contents=contents,
                    )
                    events.append(FileUpdate(
                        path=holograph_path,
                        contents=contents,
                    ))
                else:
                    # this is a new file, create a new entry in the file map
                    last_updated = os.path.getmtime(abs_path)
                    with open(abs_path) as h:
                        contents = h.read()
                    files_map[holograph_path] = File(
                        path=holograph_path,
                        last_modified=last_updated,
                        contents=contents,
                    )
                    events.append(FileAdd(
                        path=holograph_path,
                        contents=contents,
                    ))
        # now check which files have been deleted
        deleted_file_paths = old_files_set - new_files_set
        for deleted_path in deleted_file_paths:
            del files_map[deleted_path]
            events.append(FileDelete(path=deleted_path))
        old_files_set = new_files_set
        if len(events) > 0:
            emit_events(url, channel_id, events, encryptor)
        time.sleep(poll_interval_ms / 1000)

def emit_events(url: str, channel_id: str, events: List[Union[FileAdd, FileUpdate, FileDelete]], encryptor: AESOCB3):
    event_objs = [e.to_channel_event(encryptor).to_dict() for e in events]
    response = requests.post(f"{url}/channel/{channel_id}", json=event_objs)
    response.raise_for_status()

def open_channel(url: str) -> str:
    response = requests.post(f"{url}/channel", json={})
    response.raise_for_status()
    return response.text

def close_channel(url: str, channel: str):
    response = requests.delete(f"{url}/channel/{channel}")
    response.raise_for_status()

def print_channel_details(url: str, channel: str, key: bytes):
    """
    Prints the channel details in a format that can simply be copied into the 
    target python script.
    """
    encoded_key = base64.b64encode(key).decode()
    code_snippet = textwrap.dedent(f'''
    
    import holograph

    holograph.tune_in(
        key="{encoded_key}",
        url="{url}",
        channel="{channel}",
    )

    ''')
    console = Console()
    intro_message = Text("Tune in from your remote location using the following code which ", style="bold")
    intro_message.append("must be kept secret for the duration of your session", style="bold red")
    intro_message.append(":")
    console.print(intro_message)
    syntax = Syntax(code_snippet, "python", background_color="default")
    console.print(syntax)
    console.print("Transmitting file updates... ", style="bold")

if __name__ == "__main__":
    main()


import atexit
import base64
import importlib.abc
from importlib.machinery import (
    ModuleSpec,
    ExtensionFileLoader,
    SourceFileLoader,
    SourcelessFileLoader,
    SOURCE_SUFFIXES,
    BYTECODE_SUFFIXES,
    EXTENSION_SUFFIXES,
)
from importlib.util import spec_from_file_location
import os
import shutil
import sys
import tempfile
from typing import List
import warnings

from cryptography.hazmat.primitives.ciphers.aead import AESOCB3
from cryptography.exceptions import InvalidTag
import requests

from .models import (
    ChannelEvent,
    EventAction,
)


HOLOGRAPH_MAGIC_PATH_ENTRY = "HOLOGAPH_REMOTE_LOCATION"
TEMP_MAGIC_PATH = "/Users/max.taggart/Developer/Holograph/magic_location"

def tune_in(url: str, channel: str, key: str):
    """
    This function is the entrypoint on the consuming-side of a holograph transmission.
    i.e. users call this function where they wish to consume the code that they are
    writing in a more authoring-friendly environment.
    It hooks into the native python import logic to enable imports from the
    Holograph server.
    """
    key_bytes = base64.b64decode(key.encode())
    # Instantiate the same file loaders that are used by the built-in FileFinder
    extensions = (ExtensionFileLoader, EXTENSION_SUFFIXES)
    source = (SourceFileLoader, SOURCE_SUFFIXES)
    bytecode = (SourcelessFileLoader, BYTECODE_SUFFIXES)
    loader_details = [extensions, source, bytecode]
    # Create the temporary directory where we can store the remote files, this allows
    # us to re-use a lot of the existing logic from the standard FileFinder around
    # caching file contents.
    temp_dir = tempfile.mkdtemp(prefix="holograph-")
    print(f"Temp Dir: {temp_dir}")
    atexit.register(lambda: shutil.rmtree(temp_dir))
    # Add the HolographicFileFinder to the list of path_hooks so that the PathFinder knows about it.
    sys.path_hooks.append(HolographicFileFinder.path_hook(url, channel, key_bytes, temp_dir, *loader_details))
    # Add the HOLOGRAPH_MAGIC_PATH_ENTRY to sys.path. Our HolographicFileFinder will only be asked to
    # find modules if there is an entry in sys.path which is not actually a directory. Otherwise, if the
    # entry in sys.path is a real directory PathFinder will just create a new instance of the regular 
    # FileFinder to handle it, which of course won't know how to communicate with the Holograph server. 
    # But, since HolographicFileFinder knows to look for the HOLOGRAPH_MAGIC_PATH_ENTRY in the module's path
    # its `path_hook` classmethod will return an instance of HolographicFileFinder when it sees the magic 
    # path, that finder will be added to the `sys.path_importer_cache` for the path starting with
    # `HOLOGRAPH_MAGIC_PATH_ENTRY`, and we will have successfully hooked into the existing PathFinder logic.
    sys.path.append(HOLOGRAPH_MAGIC_PATH_ENTRY)
    
class DataIntegrityError(Exception):
    """
    Raised when we encounter an invalid signature for data coming from the holograph server.
    """
    pass

class HolographicFileFinder(importlib.abc.PathEntryFinder):

    """
    This class is a modified version of the native imporlib.machinery.FileFinder which
    knows how to ask the Holograph server for updates before determining whether or not
    it can find the requested module. It also knows to look for module paths that start
    with the `HOLOGRAPH_MAGIC_PATH_ENTRY` and manipulates the module path in the 
    ModuleSpec to start with `HOLOGRAPH_MAGIC_PATH_ENTRY` so that sub-modules are also
    imported using an instance of this class.
    """

    def __init__(self, path: str, url, channel, key, temp_dir, *loader_details):
        """Initialize with the path to search on and a variable number of
        2-tuples containing the loader and the file suffixes the loader
        recognizes."""
        loaders = []
        for loader, suffixes in loader_details:
            loaders.extend((suffix, loader) for suffix in suffixes)
        self._loaders = loaders
        # The `path` variable passed to us will start with `HOLOGRAPH_MAGIC_PATH_ENTRY`, 
        # but we need to convert that into an actual path that starts with `temp_dir`.
        path_end = path.replace(HOLOGRAPH_MAGIC_PATH_ENTRY, "")
        # Note that if path_end starts with a "/" then os.path.join will just return path_end
        # since it interprets it as an absolute path.
        if len(path_end) > 0 and path_end[0] == "/":
            path_end = path_end[1:]
        self.path = os.path.join(temp_dir, path_end)
        self._path_mtime = -1
        self._path_cache = set()
        self._relaxed_path_cache = set()
        self._url = url
        self._channel = channel
        self._decryptor = AESOCB3(key)
        self._temp_dir = temp_dir

    def invalidate_caches(self):
        """Invalidate the directory mtime."""
        self._path_mtime = -1

    def find_module(self, fullname):
        """Try to find a loader for the specified module by delegating to
        self.find_loader().

        This method is deprecated in favor of finder.find_spec().

        """
        warnings.warn("find_module() is deprecated and "
                    "slated for removal in Python 3.12; use find_spec() instead",
                    DeprecationWarning)
        # Call find_loader(). If it returns a string (indicating this
        # is a namespace package portion), generate a warning and
        # return None.
        loader, portions = self.find_loader(fullname)
        if loader is None and len(portions):
            msg = 'Not importing directory {}: missing __init__'
            warnings.warn(msg.format(portions[0]), ImportWarning)
        return loader

    def find_loader(self, fullname):
        """Try to find a loader for the specified module, or the namespace
        package portions. Returns (loader, list-of-portions).

        This method is deprecated.  Use find_spec() instead.

        """
        warnings.warn("FileFinder.find_loader() is deprecated and "
                       "slated for removal in Python 3.12; use find_spec() instead",
                       DeprecationWarning)
        spec = self.find_spec(fullname)
        if spec is None:
            return None, []
        return spec.loader, spec.submodule_search_locations or []

    def _get_spec(self, loader_class, fullname, path, smsl, target):
        loader = loader_class(fullname, path)
        module_spec = spec_from_file_location(fullname, path, loader=loader,
                                       submodule_search_locations=smsl)
        if module_spec is None:
            return None
        # If we found the module then make sure to overwrite its __path__ so that it starts
        # with our magic path entry, otherwise if this module is really a package then any
        # sub-packages will be imported using the actual path, which will cause the MetaPathFinder
        # to use a regular FileFinder, which will be able to find the file, but won't trigger
        # a call to the server to update the local filesystem with remote changes.
        # According to the importlib documentation for ModuleSpecs
        # (https://docs.python.org/3/library/importlib.html#importlib.machinery.ModuleSpec) 
        # the `submodule_search_locations` attribute is what is used for the module's __path__
        # attribute, which is what is passed to the MetaPathFinder when importing a sub-module
        # from a package. Quoting the docs, this attribute is
        # 
        # "The list of locations where the package’s submodules will be found. Most of the time this 
        # is a single directory. The finder should set this attribute to a list, even an empty one, 
        # to indicate to the import system that the module is a package. It should be set to None 
        # for non-package modules. It is set automatically later to a special object for namespace 
        # packages.""
        # 
        # So if it is already set to None we will just leave it as is, but if it is a list then we
        # know that the module is actually a package, in which case we need to overwrite the entries
        # in the list so they start with our magic location.
        if module_spec.submodule_search_locations is None:
            return module_spec
        # Make sure the temp_dir does not end in a slash so that when we replace the temp_dir
        # name with our magic path entry we don't pull out the slash that we need to separate 
        # the magic path entry from the first actual filename or directory.
        temp_dir = self._temp_dir
        if temp_dir[-1] == "/":
            temp_dir = temp_dir[:-1]
        for i, location in enumerate(module_spec.submodule_search_locations):
            if self._temp_dir in location:
                module_spec.submodule_search_locations[i] = location.replace(self._temp_dir, HOLOGRAPH_MAGIC_PATH_ENTRY)
        return module_spec

    def find_spec(self, fullname, target=None):
        """Try to find a spec for the specified module.

        Returns the matching spec, or None if not found.
        """
        self._apply_remote_updates()
        is_namespace = False
        tail_module = fullname.rpartition('.')[2]
        try:
            mtime = os.stat(self.path or os.getcwd()).st_mtime
        except OSError:
            mtime = -1
        if mtime != self._path_mtime:
            self._fill_cache()
            self._path_mtime = mtime
        # # tail_module keeps the original casing, for __file__ and friends
        # if _relax_case():
        #     cache = self._relaxed_path_cache
        #     cache_module = tail_module.lower()
        # else:
        #     cache = self._path_cache
        #     cache_module = tail_module
        
        cache = self._path_cache
        cache_module = tail_module
        # Check if the module is the name of a directory (and thus a package).
        # This line is here to help keep the typechecker happy
        base_path = ""
        if cache_module in cache:
            base_path = os.path.join(self.path, tail_module)
            for suffix, loader_class in self._loaders:
                init_filename = '__init__' + suffix
                full_path = os.path.join(base_path, init_filename)
                if os.path.isfile(full_path):
                    return self._get_spec(loader_class, fullname, full_path, [base_path], target)
            else:
                # If a namespace package, return the path if we don't
                #  find a module in the next section.
                is_namespace = os.path.isdir(base_path)
        # Check for a file w/ a proper suffix exists.
        for suffix, loader_class in self._loaders:
            try:
                full_path = os.path.join(self.path, tail_module + suffix)
            except ValueError:
                return None
            # _verbose_message('trying {}', full_path, verbosity=2)
            if cache_module + suffix in cache:
                if os.path.isfile(full_path):
                    return self._get_spec(loader_class, fullname, full_path,
                                          None, target)
        if is_namespace:
            # _verbose_message('possible namespace for {}', base_path)
            spec = ModuleSpec(fullname, None)
            spec.submodule_search_locations = [base_path]
            return spec
        return None

    def _fill_cache(self):
        """Fill the cache of potential modules and packages for this directory."""
        path = self.path
        try:
            contents = os.listdir(path or os.getcwd())
        except (FileNotFoundError, PermissionError, NotADirectoryError):
            # Directory has either been removed, turned into a file, or made
            # unreadable.
            contents = []
        # We store two cached versions, to handle runtime changes of the
        # PYTHONCASEOK environment variable.
        if not sys.platform.startswith('win'):
            self._path_cache = set(contents)
        else:
            # Windows users can import modules with case-insensitive file
            # suffixes (for legacy reasons). Make the suffix lowercase here
            # so it's done once instead of for every import. This is safe as
            # the specified suffixes to check against are always specified in a
            # case-sensitive manner.
            lower_suffix_contents = set()
            for item in contents:
                name, dot, suffix = item.partition('.')
                if dot:
                    new_name = '{}.{}'.format(name, suffix.lower())
                else:
                    new_name = name
                lower_suffix_contents.add(new_name)
            self._path_cache = lower_suffix_contents
        # if sys.platform.startswith(_CASE_INSENSITIVE_PLATFORMS):
        #     self._relaxed_path_cache = {fn.lower() for fn in contents}

    def _apply_remote_updates(self):
        """
        Get any updates for this channel from the Holograph server, including
        decrypting and checking the signature of the updates.
        """
        url = f"{self._url}/channel/{self._channel}/updates"
        response = requests.get(url)
        response.raise_for_status()
        body_json: List[dict] = response.json()
        for d in body_json:
            event = ChannelEvent.from_dict(d)
            # Make sure the signature is valid
            try:
                file_contents = self._decryptor.decrypt(event.nonce, event.contents, event.filepath.encode())
            except InvalidTag:
                # Either the cyphertext or the path changed
                raise DataIntegrityError(
                    "One of the source files retrieved from the Holograph server has been "
                    "altered en-route after leaving the source filesystem. "
                    "Please open a new channel and re-initialize your Holograph connection."
                    f"(Suspicious file: {event.filepath})"
                )
            filepath = os.path.join(self._temp_dir, event.filepath)
            if event.action in [EventAction.CREATE, EventAction.UPDATE]:
                # We handle CREATE and UPDATE events the same because the server will send as an UPDATE event
                # for a file we have never seen before if that file was both created and updated since the last
                # time we asked for udpates (similar to the case where a file is both created and deleted in 
                # between udpates as explained below).
                # Make sure the intermediate directories exist first
                os.makedirs(os.path.dirname(filepath), exist_ok=True)
                # Then write the file
                with open(filepath, 'wb') as h:
                    h.write(file_contents)
            elif event.action == EventAction.DELETE:
                # The Holograph server does not keep track of the state of the filesystem on the consuming end.
                # This means that if a file on the producing end was both created and deleted since we last asked 
                # for updates from the server then the server will send us a "DELETE" event for that file, even
                # though the file was never present on the consuming filesystem. So we simply need to check for
                # the existence of the file before asking the operating system to delete it in case we are in
                # this situation.
                if os.path.exists(filepath):
                    os.remove(filepath)




    @classmethod
    def path_hook(cls, url, channel, key, temp_dir, *loader_details):
        """A class method which returns a closure to use on sys.path_hook
        which will return an instance using the specified loaders and the path
        called on the closure.

        If the path called on the closure is not a directory, ImportError is
        raised.

        """
        def path_hook_for_HolographicFileFinder(path: str):
            """Path hook for for the file finder used by Holograph."""
            # We hook into the import machinery by adding a dummy path to sys.path
            # which we look for here. The first time the import machinery looks
            # for a module that is at the top-level of our remote package it will 
            # pass `path == HOLOGRAPHIC_MAGIC_PATH_ENTRY`, which will be our queue
            # to return an instance of this class which knows how to look for modules
            # that are coming from the Holograph server. For child modules the path
            # will start with `HOLOGRAPHIC_MAGIC_PATH_ENTRY`.
            if not path.startswith(HOLOGRAPH_MAGIC_PATH_ENTRY):
                raise ImportError(f'the HolographicFileFinder only handles modules at the magic "{HOLOGRAPH_MAGIC_PATH_ENTRY}" path.')
            return cls(path, url, channel, key, temp_dir, *loader_details)

        return path_hook_for_HolographicFileFinder

    def __repr__(self):
        return 'HolographicFileFinder({!r})'.format(self.path)


# # Taken from Lib/importlib/_bootstrap_external.py:54 (branch 3.10)
# _CASE_INSENSITIVE_PLATFORMS_STR_KEY = 'win',
# _CASE_INSENSITIVE_PLATFORMS_BYTES_KEY = 'cygwin', 'darwin'
# _CASE_INSENSITIVE_PLATFORMS =  (_CASE_INSENSITIVE_PLATFORMS_BYTES_KEY
#                                 + _CASE_INSENSITIVE_PLATFORMS_STR_KEY)
# def _make_relax_case():
#     if sys.platform.startswith(_CASE_INSENSITIVE_PLATFORMS):
#         if sys.platform.startswith(_CASE_INSENSITIVE_PLATFORMS_STR_KEY):
#             key = 'PYTHONCASEOK'
#         else:
#             key = b'PYTHONCASEOK'

#         def _relax_case():
#             """True if filenames must be checked case-insensitively and ignore environment flags are not set."""
#             return not sys.flags.ignore_environment and key in os.environ
#     else:
#         def _relax_case():
#             """True if filenames must be checked case-insensitively."""
#             return False
#     return _relax_case

# _relax_case = _make_relax_case()

# # Taken from Lib/importlib/_bootstrap.py:244 (branch 3.10)
# def _verbose_message(message, *args, verbosity=1):
#     """Print the message to stderr if -v/PYTHONVERBOSE is turned on."""
#     if sys.flags.verbose >= verbosity:
#         if not message.startswith(('#', 'import ')):
#             message = '# ' + message
#         print(message.format(*args), file=sys.stderr)